#include <libyang/libyang.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <libssh/libssh.h>
#include <libyang/libyang.h>
#include <nc_server.h>

static volatile int loop_continue = 1;

static void signal_handler(int sig)
{
  switch (sig) {
    case SIGINT:
    case SIGTERM:
    case SIGQUIT:
    case SIGABRT:
    case SIGHUP:
        loop_continue = 0;
        break;
    default:
        exit(EXIT_FAILURE);
    }
}

static int interactive_clb(const struct nc_session *session, ssh_session ssh_session, ssh_message msg, void *) {
  const char *name = "Keyboard-Interactive Authentication";
  char *prompt = "Password: ";
  char *received_pw = NULL;
  char echo[] = {0};
  const char *username = ssh_message_auth_user(msg);

  nc_libssh_thread_verbosity(3);
  ssh_set_log_level(3);

  /* send the password prompt to the client */
  int ret = ssh_message_auth_interactive_request(msg, name, "", 1, (const char **)&prompt, echo);
  if (ret) {
    printf("Failed to send an authentication request to client: %s\n", username);
    return ret;
  }

  /* get the reply */
  int n_answers = nc_server_ssh_kbdint_get_nanswers(session, ssh_session);
  if (n_answers != 1) {
    /* only expecting a single answer */
    printf("Unexpected amount of answers in interactive auth. Expected 1, got: %d\n", n_answers);
    return -1;
  }

  received_pw = strdup(ssh_userauth_kbdint_getanswer(ssh_session, 0));
  
  printf("Received password: %s\n", received_pw);

  return strcmp(received_pw, "test");
}

void sshlog_clb(int priority, const char *function, const char* buffer, void *userdata) {
  printf("SSH log: %s: %s\n", function, buffer);
}

int main() {
  int ret;
  struct sigaction action;
  sigset_t block_mask;
  struct ly_ctx *ctx = NULL;
  struct nc_session *ncs = NULL;
  struct lyd_node *server_config = NULL;
  NC_MSG_TYPE msgtype;

  sigfillset(&block_mask);
  action.sa_handler = signal_handler;
  action.sa_mask = block_mask;
  action.sa_flags = 0;
  sigaction(SIGINT, &action, NULL);
  sigaction(SIGQUIT, &action, NULL);
  sigaction(SIGABRT, &action, NULL);
  sigaction(SIGTERM, &action, NULL);
  sigaction(SIGHUP, &action, NULL);

  nc_verbosity(NC_VERB_DEBUG);
  ly_log_level(LY_LLDBG);
  ssh_set_log_callback(sshlog_clb);

  if (ly_ctx_new(NULL, 0, &ctx) != LY_SUCCESS) {
    printf("ly_ctx_new_error()!\n");
  }

  if (ly_ctx_set_searchdir(ctx, "./yang_modules") != LY_SUCCESS) {
    printf("Could not set searchdir!\n");
  }

  nc_server_init();
  if (nc_server_init_ctx(&ctx) != 0) {
    printf("nc_server_init_ctx() error!\n");
  }

  if (nc_server_config_load_modules(&ctx) != 0) {
    printf("nc_server_config_load_modules() error!\n");
  }

  // SSH config
  nc_server_ssh_set_interactive_auth_clb(interactive_clb, NULL, NULL);
  ret = nc_server_config_add_address_port(ctx, "ssh_endpt", NC_TI_LIBSSH, "0.0.0.0", 10830, &server_config);
  if (ret != 0) {
    printf("nc_server_config_add_address_port failed!\n");
    goto cleanup;
  }

  ret = nc_server_config_add_ssh_hostkey(ctx, "ssh_endpt", "id_rsa", "./test_keys", "./test_keys.pub", &server_config);
  if (ret != 0) {
    printf("nc_server_config_add_ssh_hostkey failed!\n");
    goto cleanup;
  }

  // Left this commented out intentionally. Don't support password auth
  //ret = nc_server_config_add_ssh_user_password(ctx, "ssh_endpt", "test", "", &server_config);

  ret = lyd_new_path(server_config, ctx, "/ietf-netconf-server:netconf-server/listen/endpoints/endpoint[name='ssh_endpt']/ssh/"
            "ssh-server-parameters/client-authentication/users/user[name='test']/libnetconf2-netconf-server:keyboard-interactive/use-system-auth", NULL, 0, NULL);
  if (ret != 0) {
    printf("lyd_new_path failed!\n");
    goto cleanup;
  }

  ret = nc_server_config_setup_data(server_config);
  if (ret != 0) {
    printf("nc_server_config_setup_data failed!\n");
    goto cleanup;
  }

  while (loop_continue) {
    msgtype = nc_accept(0 ,ctx, &ncs);
    if (msgtype == NC_MSG_HELLO) {
      printf("HELLO received\n");
    }
    nc_session_free(ncs, NULL);
  }

cleanup:
  ly_ctx_destroy(ctx);
  nc_server_destroy();

  return 0;
}
